import json
from datetime import datetime
from typing import Optional, TYPE_CHECKING, List

from boto3.dynamodb.conditions import Key

from warehouse14 import Account, Token, Project
from warehouse14.models import Version
from warehouse14.repos import DBBackend

if TYPE_CHECKING:
    from mypy_boto3_dynamodb.service_resource import Table, DynamoDBServiceResource


def create_table(dynamodb: "DynamoDBServiceResource", table_name):
    if table_name in [table.name for table in dynamodb.tables.all()]:
        return dynamodb.Table(table_name)

    return dynamodb.create_table(
        TableName=table_name,
        AttributeDefinitions=[
            {"AttributeName": "pk", "AttributeType": "S"},
            {"AttributeName": "sk", "AttributeType": "S"},
        ],
        KeySchema=[
            {"AttributeName": "pk", "KeyType": "HASH"},
            {"AttributeName": "sk", "KeyType": "RANGE"},
        ],
        BillingMode="PAY_PER_REQUEST",
        GlobalSecondaryIndexes=[
            {
                "IndexName": "sk_gsi",
                "KeySchema": [
                    {"AttributeName": "sk", "KeyType": "HASH"},
                    {"AttributeName": "pk", "KeyType": "RANGE"},
                ],
                "Projection": {
                    "ProjectionType": "ALL",
                },
            }
        ],
    )


class DynamoDBBackend(DBBackend):
    def __init__(self, table: "Table"):
        self._table = table

    # account methods
    def account_save(self, user_id: str, **kwargs) -> Optional[Account]:
        """
        Load non confidential data for a user
        """
        # check if already exists, if so, get created date
        existing_account = self.account_get(user_id=user_id)
        created = existing_account.created if existing_account else datetime.now()

        self._table.put_item(
            Item={
                "pk": f"account#{user_id}",
                "sk": f"account#{user_id}",
                "created": created.isoformat(),
                "updated": datetime.now().isoformat(),
            }
        )
        return self.account_get(user_id=user_id)

    def account_get(self, user_id: str) -> Optional[Account]:
        """
        Load non confidation data for a user
        :param user_id: ID of the user, which will be used to search for
        """

        item = self._table.get_item(
            Key={
                "pk": f"account#{user_id}",
                "sk": f"account#{user_id}",
            }
        ).get("Item")

        if item is None:
            return None
        else:
            return Account(
                name=user_id,
                created=datetime.fromisoformat(item.get("created")),
                updated=datetime.fromisoformat(item.get("updated")),
            )

    def account_token_add(
        self, user_id: str, token_id: str, name: str, key: str
    ) -> Token:
        now = datetime.now()
        self._table.put_item(
            Item={
                "pk": f"account#{user_id}",
                "sk": f"token#{token_id}",
                "name": name,
                "key": key,
                "created": now.isoformat(),
            }
        )
        return Token(
            id=token_id,
            name=name,
            key=key,
            created=now,
        )

    def account_token_list(self, account_id: str) -> List[Token]:
        query = Key("pk").eq(f"account#{account_id}") & Key("sk").begins_with("token")
        items = self._table.query(KeyConditionExpression=query).get("Items", [])

        return [
            Token(
                id=item["sk"].split("#")[1],
                name=item["name"],
                key=item["key"],
                created=datetime.fromisoformat(item["created"]),
            )
            for item in items
        ]

    def account_token_delete(self, account_id: str, token_id: str):
        self._table.delete_item(
            Key={
                "pk": f"account#{account_id}",
                "sk": f"token#{token_id}",
            }
        )

    def resolve_token(self, token_id: str) -> Optional[Account]:
        """
        Look up a account by given token id.

        :param token_id: Unique token id
        :return: Account
        """

        items = self._table.query(
            IndexName="sk_gsi", KeyConditionExpression=Key("sk").eq(f"token#{token_id}")
        ).get("Items", [])

        if len(items) > 1:
            raise Exception(
                f"Not able to resolve token, found to many ({len(items)}) accounts."
            )
        elif len(items) == 0:
            return None

        account_id = items[0]["pk"].split("#")[1]
        return self.account_get(account_id)

    # Project methods
    def project_save(self, project: Project) -> Project:
        normalized_name = project.normalized_name()

        current = self.project_get(normalized_name)

        # calc admins diff
        current_admins = set(current.admins) if current else set()
        project_admins = set(project.admins)
        to_create_admins = project_admins - current_admins
        to_delete_admins = current_admins - project_admins

        # calc members diff
        current_members = set(current.members) if current else set()
        project_members = set(project.members)
        to_create_members = project_members - current_members
        to_delete_members = current_members - project_members

        # write to db
        with self._table.batch_writer() as w:
            # update project entry
            w.put_item(
                Item={
                    "pk": f"project#{normalized_name}",
                    "sk": f"project#{normalized_name}",
                    "name": project.name,
                    "versions": {
                        k: json.loads(v.json()) for k, v in project.versions.items()
                    },
                }
            )

            # update public state
            if project.public:
                w.put_item(
                    Item={
                        "pk": f"project#{normalized_name}",
                        "sk": f"account#public",
                        "role": "member",
                    }
                )
            else:
                w.delete_item(
                    Key={
                        "pk": f"project#{normalized_name}",
                        "sk": f"account#public",
                    }
                )

            # write admin diff
            for admin in to_create_admins:
                w.put_item(
                    Item={
                        "pk": f"project#{normalized_name}",
                        "sk": f"account#{admin}",
                        "name": admin,
                        "role": "admin",
                    }
                )
            for admin in to_delete_admins:
                w.delete_item(
                    Key={
                        "pk": f"project#{normalized_name}",
                        "sk": f"account#{admin}",
                    }
                )

            # write member diff
            for member in to_create_members:
                w.put_item(
                    Item={
                        "pk": f"project#{normalized_name}",
                        "sk": f"account#{member}",
                        "name": member,
                        "role": "member",
                    }
                )
            for member in to_delete_members:
                w.delete_item(
                    Key={
                        "pk": f"project#{normalized_name}",
                        "sk": f"account#{member}",
                    }
                )

        return self.project_get(normalized_name)

    def project_get(self, name: str) -> Optional[Project]:
        normalized_name=Project.normalize_name(name)

        items = self._table.query(
            KeyConditionExpression=Key("pk").eq(f"project#{normalized_name}")
        ).get("Items", [])

        db_project = None
        public = False
        db_admins = []
        db_members = []
        for item in items:
            if item["sk"].startswith("project#"):
                db_project = item
            if item["sk"] == "account#public":
                public = True
            elif item["sk"].startswith("account#"):
                role = item.get("role")
                if role == "admin":
                    db_admins.append(item)
                elif role == "member":
                    db_members.append(item)

        if db_project is None:
            return None

        return Project(
            name=db_project["name"],
            admins=[a["name"] for a in db_admins],
            members=[a["name"] for a in db_members],
            public=public,
            versions={k: Version(**v) for k, v in db_project["versions"].items()},
        )

    def project_list(self) -> List[Project]:
        """
        Lists all project names
        """
        # TODO this will explode,
        # we have to scan the whole DB and do additional requests per project
        # we should lookup by the index for a specific user, using a paginator!
        items = self._table.scan(
            FilterExpression=Key("pk").begins_with(f"project#")
            & Key("sk").begins_with("project#"),
        ).get("Items", [])

        return [self.project_get(p["name"]) for p in items]



import datetime
import secrets
from collections import defaultdict
from operator import attrgetter
from uuid import uuid4

import pypitoken
import readme_renderer.markdown
import readme_renderer.rst
import readme_renderer.txt
from flask import Flask, render_template, redirect, url_for, abort, request, g
from flask_login import LoginManager, login_required, current_user, logout_user
from flaskext.markdown import Markdown

from warehouse14 import simple_api
from warehouse14.forms import CreateProjectForm, CreateAPITokenForm
from warehouse14.login import OIDCAuthenticator, Authenticator, User
from warehouse14.models import Project, Account, Token
from warehouse14.pkg_helpers import guess_pkgname_and_version
from warehouse14.repos import DBBackend
from warehouse14.storage import SimpleFileStorage, PackageStorage

_RENDERERS = {
    None: readme_renderer.rst,  # Default if description_content_type is None
    "": readme_renderer.rst,  # Default if description_content_type is None
    "text/plain": readme_renderer.txt,
    "text/x-rst": readme_renderer.rst,
    "text/markdown": readme_renderer.markdown,
}


def create_app(
    db: DBBackend, storage: PackageStorage, auth: Authenticator, config: dict = None, allow_project_creation=False
):
    app = Flask(__name__)
    if config:
        app.config.update(**config)

    Markdown(app, extensions=["footnotes", "fenced_code"])

    # Setup Login and authentication
    login_manager = LoginManager(app)

    @login_manager.user_loader
    def load_user(user_id):
        """Automatically creates a user for any OIDC login"""
        account = db.account_get(user_id)
        if account is None:
            account = db.account_save(user_id)
        return User(account) if account else None

    auth.init_app(app)

    simple_blueprint = simple_api.create_blueprint(
        db, storage, allow_project_creation=allow_project_creation
    )
    app.register_blueprint(simple_blueprint)

    def get_user_id():
        return current_user.account.name

    @app.route("/logout")
    @login_required
    def logout():
        logout_user()
        return auth.logout()

    @app.template_filter("sn")
    def filter_suppress_none(val):
        if val is not None:
            return val
        else:
            return ""

    @app.get("/")
    @login_required
    def index():
        return render_template("home.html")

    @app.get("/manage/account")
    @login_required
    def account():
        token_list = [
            (token.id, token.name) for token in db.account_token_list(get_user_id())
        ]
        return render_template("account/account.html", tokens=token_list)

    @app.route("/manage/account/token", methods=["GET", "POST"])
    @login_required
    def account_token():
        form = CreateAPITokenForm()
        if form.is_submitted():
            # create new token
            token_name = form.name.data
            token = Token(
                id=str(uuid4()),
                name=token_name,
                key=secrets.token_hex(16),
                created=datetime.datetime.utcnow(),
            )

            pypi_token = pypitoken.Token.create(
                domain="warehouse14",
                identifier=token.id,
                key=token.key,
                prefix="wh14",
            )

            # save token
            db.account_token_add(
                user_id=get_user_id(), token_id=token.id, name=token.name, key=token.key
            )

            # render with token view
            return render_template(
                "account/token.html",
                create_form=CreateAPITokenForm(formdata=None),
                new_token=token,
                token_raw=pypi_token.dump(),
            )

        return render_template("account/token.html", create_form=form)

    @app.get("/manage/account/token/delete")
    @login_required
    def account_token_delete():
        token_id = request.args.get("token_id")
        db.account_token_delete(get_user_id(), token_id)
        return redirect(url_for("account"))

    @app.get("/projects")
    @login_required
    def list_projects():
        own_projects = [
            project for project in db.project_list() if project.visible(get_user_id())
        ]
        own_projects.sort(key=attrgetter("name"))

        return render_template("project/projects.html", projects=own_projects)

    @app.get("/projects/<project_name>")
    @login_required
    def show_project(project_name):
        project = db.project_get(project_name)
        if project is None:
            abort(404)

        # In case no version is available, nothing has to be prepared
        latest_version = project.latest_version
        if latest_version is None:
            return render_template("project/project.html", project=project)

        metadata = latest_version.metadata
        # Render description depending on the type (plain, rst, markdown)
        renderer = _RENDERERS[latest_version.description_content_type]
        readme = renderer.render(latest_version.description)

        grouped_classifiers = defaultdict(list)
        for classifier in latest_version.classifiers:
            group, value = classifier.split(" :: ", 1)
            grouped_classifiers[group].append(value)

        # Parse URLS
        urls = []
        if url := metadata.get("url"):
            urls.append(("Homepage", url))
        for project_url in metadata.get("project_urls", []):

            if "=" in project_url:
                name, url = project_url.split("=", 1)
            elif "," in project_url:
                name, url = project_url.split(",", 1)
            else:
                name, url = ("URL", project_url)
            urls.append((name.strip(), url.strip()))

        return render_template(
            "project/project.html",
            project=project,
            readme=readme,
            grouped_classifiers=grouped_classifiers,
            urls=urls,
        )

    @app.route("/create_project", methods=["GET", "POST"])
    @login_required
    def create_project():
        form = CreateProjectForm()
        if form.validate_on_submit():
            project = db.project_save(
                Project(
                    name=form.name.data,
                    admins=[get_user_id()],
                    members=[],
                    public=form.public.data,
                )
            )

            return redirect(
                url_for("show_project", project_name=project.normalized_name())
            )
        else:
            return render_template("project/create_project.html", form=form)

    @app.route("/projects/<project_name>/edit", methods=["GET", "POST"])
    @login_required
    def edit_project(project_name):
        project = db.project_get(project_name)
        if not project:
            abort(404, f"No Project found with name {project_name}")

        form = CreateProjectForm(name=project.name, public=project.public)
        form.name.render_kw = {"disabled": "disabled"}

        if form.validate_on_submit():
            # Update fields
            project.public = form.public.data

            return redirect(url_for("show_project", project_name=form.name.data))
        else:
            return render_template(
                "project/project_edit.html", project=project, form=form
            )

    @app.get("/projects/<project_name>/users")
    @login_required
    def project_users(project_name):
        project = db.project_get(project_name)
        if not project:
            abort(404, f"No Project found with name {project_name}")

        return render_template("project/project_users.html", project=project)

    @app.post("/projects/<project_name>/admins")
    @login_required
    def project_add_admin(project_name):
        print(request.form["username"])
        project = db.project_get(project_name)
        project.admins.append(request.form["username"])

        db.project_save(project)
        return redirect(url_for("project_users", project_name=project_name))

    @app.get("/projects/<project_name>/admins/<username>")
    @login_required
    def project_remove_admin(project_name, username):
        project = db.project_get(project_name)

        # Never delete the last admin!
        if len(project.admins) > 1 and username in project.admins:
            project.admins.remove(username)
        db.project_save(project)
        return redirect(url_for("project_users", project_name=project_name))

    @app.post("/projects/<project_name>/members")
    @login_required
    def project_add_member(project_name):
        print(request.form["username"])
        project = db.project_get(project_name)
        project.members.append(request.form["username"])
        db.project_save(project)
        return redirect(url_for("project_users", project_name=project_name))

    @app.get("/projects/<project_name>/users/<username>")
    @login_required
    def project_remove_member(project_name, username):
        project = db.project_get(project_name)
        if username in project.members:
            project.members.remove(username)
        db.project_save(project)
        return redirect(url_for("project_users", project_name=project_name))

    return app



from abc import abstractmethod, ABC
from typing import List, Optional

from warehouse14.models import Project, Account, Token


class DBBackend(ABC):

    # Account methods
    @abstractmethod
    def account_save(self, user_id: str, **kwargs) -> Optional[Account]:
        """
        Load non confidential data for a user
        :param user_id: ID of the user, who is represented by a account
        :param kwargs: further properties of the user
        """
        raise NotImplementedError()

    @abstractmethod
    def account_get(self, user_id: str) -> Optional[Account]:
        """
        Load non confidation data for a user
        :param user_id: ID of the user, which will be used to search for
        """
        raise NotImplementedError()

    @abstractmethod
    def account_token_add(
        self, user_id: str, token_id: str, name: str, key: str
    ) -> Optional[Token]:
        raise NotImplementedError()

    @abstractmethod
    def account_token_list(self, account_id: str) -> List[Token]:
        raise NotImplementedError()

    @abstractmethod
    def account_token_delete(self, account_id: str, token_id: str):
        raise NotImplementedError()

    @abstractmethod
    def resolve_token(self, token_id: str) -> Optional[Account]:
        """
        Look up a account by given token id.

        :param token_id: Unique token id
        :return: user_id
        """
        raise NotImplementedError

    # Project methods
    @abstractmethod
    def project_save(self, project: Project) -> Project:
        """
        Create a project in the package index
        :param project: project data
        """
        raise NotImplementedError()

    @abstractmethod
    def project_get(self, name: str) -> Optional[Project]:
        """
        Returns project by given name.

        May be overwritten by subclasses, for an optimized implementation.
        :param name: project name
        :return: Project or None
        """
        name = Project.normalize_name(name)
        for project in self.projects():
            if project.normalized_name() == name:
                return project
        else:
            return None

    @abstractmethod
    def project_list(self) -> List[Project]:
        """
        Lists all project names
        """
        raise NotImplementedError()


#
#
# class SimpleFileDB:
#     def __init__(self, root: Union[str, Path]):
#         super().__init__()
#         self._root = Path(root).expanduser().resolve()
#
#         self._project_root = self._root / "projects"
#         self._account_root = self._root / "accounts"
#
#     def _save_project(self, project: Project):
#         file = self._locate_project(project.name)
#         file.parent.mkdir(parents=True, exist_ok=True)
#
#         with file.open("wt") as f:
#             f.write(project.json())
#
#     def _locate_project(self, name: str) -> Path:
#         return self._project_root / f"{Project.normalize_name(name)}.json"
#
#     def _locate_account(self, name) -> Path:
#         return self._account_root / f"{Account.hash_name(name)}.json"
#
#     def _list_projects(self) -> List[Project]:
#         project_files = [x for x in self._project_root.glob("**/*") if x.is_file()]
#         return [self._load_project(file) for file in project_files]
#
#     def _map_token_to_account(self) -> Dict[str, Account]:
#         tokens = {}
#         account_files = [x for x in self._account_root.glob("**/*") if x.is_file()]
#         for file in account_files:
#             account = Account.parse_file(file)
#             for token in account.tokens:
#                 tokens[token.id] = account
#         return tokens
#
#     @staticmethod
#     def _load_project(file: Union[str, Path]) -> Project:
#         return Project.parse_file(Path(file))
#
#     def create_project(self, project: Project):
#         self._save_project(project)
#         return project
#
#     def update_project(self, project: Project) -> Project:
#         # TODO raise Exception
#         self._save_project(project)
#         return project
#
#     def projects(self) -> List[Project]:
#         return self._list_projects()
#
#     def delete_project(self, name: str):
#         file = self._locate_project(name)
#         file.unlink(True)
#
#     def save_account(self, account: Account) -> Optional[Account]:
#         """
#         Load non confidation data for a user
#         :param user: User object to be saved in db
#         """
#         file = self._locate_account(account.name)
#         file.parent.mkdir(parents=True, exist_ok=True)
#
#         # Store user data
#         file.write_text(account.json())
#         # TODO resolving tokens requires a different way of storage
#         # file.write_text(json.dumps({
#         #     "name": account.name
#         # }))
#
#         return account.copy()
#
#     def resolve_token(self, token_id: str) -> Optional[Account]:
#         tokens = self._map_token_to_account()
#         return tokens.get(token_id)
#
#     def load_account(self, account_id: str) -> Optional[Account]:
#         """
#         Load non confidation data for a user
#         :param user_id: ID of the user, which will be used to search for
#         """
#         file = self._locate_account(account_id)
#         if file.exists():
#             return Account.parse_file(file)
#
#         return None



from flask_wtf import FlaskForm
from wtforms import StringField, RadioField
from wtforms.validators import DataRequired


class CreateProjectForm(FlaskForm):
    name = StringField("Project Name", validators=[DataRequired()])
    public = RadioField(
        "Mode", choices=[(False, "Private"), (True, "Public")], default=False
    )


class CreateAPITokenForm(FlaskForm):
    name = StringField(validators=[DataRequired()], label="What is this token for?")



import hashlib
from abc import ABC, abstractmethod
from io import BytesIO
from pathlib import Path
from typing import Iterable, BinaryIO, Union


class PackageStorage(ABC):
    @abstractmethod
    def add(self, project: str, file: str, data: BinaryIO):
        """
        :param project: normalized name of the project
        :param file: file file of the blob to store
        :param data: blob to store
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def list(self) -> Iterable[str]:
        raise NotImplementedError()

    @abstractmethod
    def get(self, project: str, file: str) -> BinaryIO:
        raise NotImplementedError()

    @abstractmethod
    def delete(self, project: str, file: str):
        raise NotImplementedError()

    def digest(self, project: str, file: str, hash_algo: str) -> str:
        """
        Reads and digests for a file according to specified hashing-algorithm.
        :param file: file of file in storage
        :param hash_algo: any algo contained in :mod:`hashlib`
        :return: <hash_algo>=<hex_digest>

        From https://stackoverflow.com/a/21565932/548792
        """
        data = self.get(project, file)

        digester = hashlib.new(hash_algo)
        blocksize = 2 ** 16
        for block in iter(lambda: data.read(blocksize), b""):
            digester.update(block)
        return f"{hash_algo}={digester.hexdigest()}"


class SimpleFileStorage(PackageStorage):
    def __init__(self, root: Union[str, Path], allow_overwrite=False):
        self._root = (Path(root) / "packages").expanduser().resolve()
        self._allow_overwrite = allow_overwrite

    def add(self, project: str, file: str, data: BinaryIO):
        key = self._root / project / file

        if key.exists() and not self._allow_overwrite:
            raise FileExistsError(str(key))

        key.parent.mkdir(parents=True, exist_ok=True)
        key.write_bytes(data.read())

    def list(self) -> Iterable[str]:
        return [
            str(x.relative_to(self._root))
            for x in self._root.glob("**/*")
            if x.is_file()
        ]

    def get(self, project: str, file: str) -> BinaryIO:
        key = self._root / project / file
        return BytesIO(key.read_bytes())

    def delete(self, project: str, file: str):
        key = self._root / project / file
        if key.exists():
            key.unlink()


class S3Storage(PackageStorage):
    def __init__(self, bucket, allow_overwrite=False):
        self._bucket = bucket
        self._allow_overwrite = allow_overwrite

    def add(self, project: str, file: str, data: BinaryIO):
        key = f"{project}/{file}"

        if self._contains(key) and not self._allow_overwrite:
            raise FileExistsError(str(key))

        self._bucket.put_object(Key=key, Body=data.read())

    def list(self) -> Iterable[str]:
        for obj in self._bucket.objects.all():
            yield obj.key

    def get(self, project: str, file: str) -> BinaryIO:
        key = f"{project}/{file}"
        if self._contains(key):
            return self._bucket.Object(key).get()["Body"]
        else:
            raise KeyError()

    def delete(self, project: str, file: str):
        key = f"{project}/{file}"
        if self._contains(key):
            self._bucket.Object(key).delete()

    def _contains(self, key: str):
        objs = list(self._bucket.objects.filter(Prefix=key))
        return len(objs) > 0 and objs[0].key == key
